body
{
  font-family:"Palatino Linotype", "Book Antiqua", Palatino, serif;
  background-color: rgb(240, 239, 255);
  color: #333;
}

circle {
  fill: steelblue;
}

path.link {
  fill: none;
  stroke: steelblue;
}

marker {
  stroke: rgb(6,120,155);
  fill: rgb(6,120,155);
}

.interface_icon {
}

.chart div {
  font: 10px sans-serif;
  background-color: steelblue;
  text-align: right;
  margin: 1px;
  color: white;
}

.infobar {
  display: none;
  position:absolute;
  right:10px;
  width:15%;
  top: 10px;
  border: 1px solid steelblue;
  height: 90%;
  padding: 5px;
  background-color: white;
}

.infobar ul{
  list-style-type:none;
  margin-left: 0;
  padding-left: 1em;
  font-size: small;
}

.navbar {
  position:absolute;
  top: 10px;
  height: 30px;
  width: 99%;
  border-bottom: 1px solid #333;
}

.visualisation {
  position:absolute;
  top: 45px;
  width: 99%;
  height: 90%;
  background-color: rgb(240, 239, 255);
}

.chart_rect {
  fill: rgb(240, 239, 255);
}

.size_box {
  display: none;
}


